# upload_sync_controller
#
# Copyright (C) 2012-2017 Intel Corporation. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms and conditions of the GNU Lesser General Public License,
# version 2.1, as published by the Free Software Foundation.
#
# This program is distributed in the hope it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
#
# Regis Merlino <regis.merlino@intel.com>
#

import ConfigParser
import os
import shutil
import urllib
import dbus
import gio
import glib

from mediaconsole import UPNP, MediaObject, Container, Device

class _UscUpnp(UPNP):
    def __init__(self):
        UPNP.__init__(self)

    def get_servers(self):
        return self._manager.GetServers()

class _UscDevice(Device):
    def __init__(self, path):
        Device.__init__(self, path)

    def create_container_in_any(self, name, type, child_types):
        return self._deviceIF.CreateContainerInAnyContainer(name, type,
                                                            child_types)

class _UscContainer(Container):
    def __init__(self, path):
        Container.__init__(self, path)

    def create_container(self, name, type, child_types):
        return self._containerIF.CreateContainer(name, type, child_types)

    def upload(self, name, file_path):
        return self._containerIF.Upload(name, file_path)

class UscError(Exception):
    """An Upload Sync Controller error."""

    def __init__(self, message):
        """
        message: description of the error
        """

        Exception.__init__(self, message)
        self.message = message

    def __str__(self):
        return 'UscError: ' + self.message

class _UscStore(object):
    REMOTE_ID_OPTION = 'remote_id'
    TYPE_OPTION = 'type'

    def __init__(self, root_path, server_id):
        if not os.path.exists(root_path):
            os.makedirs(root_path)

        self.__config_path = os.path.join(root_path, server_id + '.conf')

        self.__config = ConfigParser.ConfigParser()

    def initialize(self):
        self.__config.read(self.__config_path)

    def __write_config(self):
        with open(self.__config_path, 'wb') as configfile:
            self.__config.write(configfile)

    def remove(self):
        if os.path.exists(self.__config_path):
            os.remove(self.__config_path)

    def __parent_deleted(self, path, deleted_containers):
        for del_path in deleted_containers:
            if path.startswith(del_path):
                return True

        return False

    def remove_file(self, path):
        id = self.__config.get(path, self.REMOTE_ID_OPTION)
        print u'\tDeleting server object {0}'.format(id)
        try:
            obj = MediaObject(id)
            obj.delete()
        except:
            pass

        print u'\tDeleting local cache {0}'.format(path)
        self.__config.remove_section(path)

    def sync_deleted_files(self, path):
        deleted_containers = []

        sections = self.__config.sections()
        if not sections:
            return

        for path in sections:
            if self.__parent_deleted(path, deleted_containers):
                print u'\tDeleting local cache {0}'.format(path)
                self.__config.remove_section(path)
            elif not os.path.exists(path):
                if self.__config.get(path, self.TYPE_OPTION) == 'container':
                    deleted_containers.append(path + os.sep)

                self.remove_file(path)

        self.__write_config()

    def add_file(self, container, name, parent):
        new_path = os.path.join(parent, name)
        if os.path.isdir(new_path):
            if self.__config.has_section(new_path):
                id = self.__config.get(new_path, self.REMOTE_ID_OPTION)
            else:
                print u'\tCreating server container for {0}'.format(new_path)
                id = container.create_container(name, 'container', ['*'])
    
                print u'\tStoring cached data for {0}'.format(id)
                self.__config.add_section(new_path)
                self.__config.set(new_path, self.REMOTE_ID_OPTION, id)
                self.__config.set(new_path, self.TYPE_OPTION, 'container')

            new_container = _UscContainer(id)
            self.sync_added_files(new_container, new_path)
        elif not self.__config.has_section(new_path):
            print u'\tUploading file {0}'.format(new_path)
            (up, id) = container.upload(name, new_path)

            print u'\tStoring cached data for {0}'.format(id)
            self.__config.add_section(new_path)
            self.__config.set(new_path, self.REMOTE_ID_OPTION, id)
            self.__config.set(new_path, self.TYPE_OPTION, 'item')

    def sync_added_files(self, container, path):
        children = os.listdir(path)
        for child in children:
            self.add_file(container, child, path)

        self.__write_config()

    def set_root(self, path, id):
        self.__config.add_section(path)
        self.__config.set(path, self.REMOTE_ID_OPTION, id)
        self.__config.set(path, self.TYPE_OPTION, 'container')

        self.__write_config()

    def object_id_from_path(self, path):
        return self.__config.get(path, self.REMOTE_ID_OPTION)

    def rename_file(self, path1, path2):
        self.__config.add_section(path2)

        id = self.__config.get(path1, self.REMOTE_ID_OPTION)
        self.__config.set(path2, self.REMOTE_ID_OPTION, id)
        type = self.__config.get(path1, self.TYPE_OPTION)
        self.__config.set(path2, self.TYPE_OPTION, type)

        self.__config.remove_section(path1)

        self.__write_config()

class UscController(object):
    """An Upload Sync Controller sample app.

    The Upload Sync Controller propagates changes in a local folder to media
    servers (DMS/M-DMS) to be added to their list of available content.
    Media servers must expose the 'content-synchronization' capability to
    be managed by this controller.

    The four main methods are servers(), track(), add_server() and start_sync().
    * servers() lists the media servers available on the network
    * track() is used to set the local folder that is to be synchronized
    * add_server() is used to add a media server to the list of servers that
      are to be synchronized
    * start_sync() launches the servers synchronisation from the local storage

    Sample usage:
    >>> controller.servers()
    >>> controller.track('/home/user/local_folder')
    >>> controller.add_server('/com/intel/dLeynaServer/server/0')
    >>> controller.start_sync()
    """

    CONFIG_PATH = os.path.join(os.environ['HOME'],
                               '.config/upload-sync-controller.conf')
    ROOT_CONTAINER_ID_OPTION = 'root_container_id'
    DATA_PATH_SECTION = '__paths__'
    DATA_PATH_OPTION = 'data_path'
    TRACK_PATH_OPTION = 'track_path'

    def __init__(self, rel_path = None):
        """
        rel_path: if provided, contains the relative local database path,
                  from the user's HOME directory.
                  If not provided, the local database path will be
                  '$HOME/upload-sync-controller'
        """

        self.__upnp = _UscUpnp()

        self.__config = ConfigParser.ConfigParser()
        self.__config.read(UscController.CONFIG_PATH)
        
        if rel_path:
            self.__set_data_path(rel_path)
        elif not self.__config.has_section(UscController.DATA_PATH_SECTION):
            self.__set_data_path('upload-sync-controller')
            
        self.__store_path = self.__config.get(UscController.DATA_PATH_SECTION,
                                              UscController.DATA_PATH_OPTION)
        try:
            self.__track_path = self.__config.get(
                                              UscController.DATA_PATH_SECTION,
                                              UscController.TRACK_PATH_OPTION)
        except:
            self.__track_path = None

    def __write_config(self):
        with open(UscController.CONFIG_PATH, 'wb') as configfile:
            self.__config.write(configfile)

    def __set_data_path(self, rel_path):
        data_path = os.environ['HOME'] + os.sep + rel_path

        if not self.__config.has_section(UscController.DATA_PATH_SECTION):
            self.__config.add_section(UscController.DATA_PATH_SECTION)

        self.__config.set(UscController.DATA_PATH_SECTION,
                          UscController.DATA_PATH_OPTION, data_path)

        self.__write_config()

    def __check_trackable(self, server):
        try:
            try:
                srt = server.get_prop('ServiceResetToken')
            except:
                raise UscError("'ServiceResetToken' variable not supported")

            try:
                dlna_caps = server.get_prop('DLNACaps')
                if not 'content-synchronization' in dlna_caps:
                    raise
            except:
                raise UscError("'content-synchronization' cap not supported")

            try:
                search_caps = server.get_prop('SearchCaps')
                if not [x for x in search_caps if 'ObjectUpdateID' in x]:
                    raise
                if not [x for x in search_caps if 'ContainerUpdateID' in x]:
                    raise
            except:
                raise UscError("'objectUpdateID' search cap not supported")

            return srt
        except UscError as err:
            print err
            return None

    def __need_sync(self, servers):
        for item in servers:
            device = _UscDevice(item)
            uuid = device.get_prop('UDN')

            if self.__config.has_section(uuid):
                yield device, uuid, self.__config.has_option(uuid,
                                    UscController.ROOT_CONTAINER_ID_OPTION)

    def __remove_monitor(self, path):
        for item in self.__monitor_list.keys():
            if item.startswith(path):
                print u'\tStop monitoring {0}'.format(path)
                del self.__monitor_list[item]

    def __add_monitor(self, path):
        print u'\tStart monitoring {0}'.format(path)
        gfile = gio.File(path)
        monitor = gfile.monitor_directory(gio.FILE_MONITOR_SEND_MOVED, None)
        monitor.connect("changed", self.__directory_changed) 
        self.__monitor_list[path] = monitor

    def __monitor_directory(self, path):
        self.__add_monitor(path)
        children = os.listdir(path)
        for child in children:
            new_path = os.path.join(path, child)
            if os.path.isdir(new_path):
                self.__monitor_directory(new_path)

    def __start_monitoring(self):
        self.__monitor_list = {}
        self.__monitor_directory(self.__track_path)

        print u'Type ctrl-c to stop.'
        try:
            main_loop = glib.MainLoop()
            main_loop.run()
        except:
            print u'Monitoring stopped.'

    def __directory_changed(self, monitor, file1, file2, evt_type):
        print
        if evt_type == gio.FILE_MONITOR_EVENT_CREATED:
            (parent, name) = os.path.split(file1.get_path())
            for store in self.__store_list:
                parent_id = store.object_id_from_path(parent)
                container = _UscContainer(parent_id)
                store.add_file(container, name, parent)

            if os.path.isdir(file1.get_path()):
                self.__monitor_directory(file1.get_path())
        elif evt_type == gio.FILE_MONITOR_EVENT_MOVED:
            (parent1, name1) = os.path.split(file1.get_path())
            (parent2, name2) = os.path.split(file2.get_path())
            renamed = (parent1 == parent2)

            if file1.get_path() in self.__monitor_list:
                self.__remove_monitor(file1.get_path())

            for store in self.__store_list:
                object_id = store.object_id_from_path(file1.get_path())
                obj = MediaObject(object_id)
                dlna_managed = obj.get_prop('DLNAManaged')

                if renamed and dlna_managed['ChangeMeta']:
                    print u'\tRenaming {0} to {1}'.format(name1, name2)

                    props = {'DisplayName' : name2}
                    obj.update(props, [])

                    store.rename_file(file1.get_path(), file2.get_path())
                else:
                    store.remove_file(file1.get_path())

                    parent_id = store.object_id_from_path(parent2)
                    container = _UscContainer(parent_id)

                    store.add_file(container, name2, parent2)

            if os.path.isdir(file2.get_path()):
                self.__monitor_directory(file2.get_path())
        elif evt_type == gio.FILE_MONITOR_EVENT_DELETED:
            if file1.get_path() in self.__monitor_list:
                self.__remove_monitor(file1.get_path())
            
            for store in self.__store_list:
                store.remove_file(file1.get_path())
        else:
            return

        print u'Type ctrl-c to stop.'

    def track(self, track_path):
        """Sets the local folder that is to be synchronized."""
        
        self.__track_path = track_path
        self.__config.set(UscController.DATA_PATH_SECTION,
                          UscController.TRACK_PATH_OPTION, track_path)

        self.__write_config()

    def start_sync(self):
        """Performs a synchronization for all the media servers.

        Displays some progress information during the process.
        """

        if not self.__track_path:
            print u'Error: track path is not set'
            return

        self.__store_list = []
        print u'Syncing...'
        for device, uuid, init in self.__need_sync(self.__upnp.get_servers()):
            store = _UscStore(self.__store_path, uuid)
            store.initialize()
            self.__store_list.append(store)

            if not init:
                print u'Performing initial sync for server {0}:'.format(uuid)
                root = device.create_container_in_any('usc_root', 'container',
                                                      ['image'])
                self.__config.set(uuid, UscController.ROOT_CONTAINER_ID_OPTION,
                                  root)
                self.__write_config()
                store.set_root(self.__track_path, root)
            else:
                print u'Performing normal sync for server {0}:'.format(uuid)
                root = self.__config.get(uuid,
                                         UscController.ROOT_CONTAINER_ID_OPTION)

            print u'\tRoot container is {0}'.format(root)
            root_container = _UscContainer(root)
            store.sync_deleted_files(self.__track_path)
            store.sync_added_files(root_container, self.__track_path)
            print u'Done.'

        if len(self.__store_list) == 0:
            print u'Nothing to do, stopping.'
            return

        print u'Now monitoring local changes...'
        self.__start_monitoring()

    def servers(self):
        """Displays media servers available on the network.

        Displays media servers information as well as the synchronized status.
        """

        print u'Running servers:'

        for item in self.__upnp.get_servers():
            try:
                server = Container(item)
                try:
                    folder_name = server.get_prop('FriendlyName')
                except Exception:
                    folder_name = server.get_prop('DisplayName')
                device = Device(item)
                dev_uuid = device.get_prop('UDN')
                dev_path = device.get_prop('Path')

                print u'{0:<25}  Synchronized({2})  {3}  {1}'.format(
                                            folder_name,
                                            dev_path,
                                            self.__config.has_section(dev_uuid),
                                            dev_uuid)

            except dbus.exceptions.DBusException as err:
                print u'Cannot retrieve properties for ' + item
                print str(err).strip()[:-1]

    def synchronized(self):
        """Displays the list of servers currently synchronized."""

        print u'Synchronized servers:'

        for name in self.__config.sections():
            if name != UscController.DATA_PATH_SECTION:
                print u'{0:<30}'.format(name)

    def add_server(self, server_path):
        """Adds a media server to the controller's list.

        server_path: d-bus path for the media server
        """

        server = Device(server_path)
        server_uuid = server.get_prop('UDN')
        
        if not self.__config.has_section(server_uuid):
            srt = self.__check_trackable(server)
            if srt != None:
                self.__config.add_section(server_uuid)
                self.__write_config()
            else:
                print u"Sorry, the server {0} has no such capability and " \
                        "will not be synchronized.".format(server_path)

    def __remove_server(self, server_uuid):
        try:
            root = self.__config.get(server_uuid,
                                     UscController.ROOT_CONTAINER_ID_OPTION)
            MediaObject(root).delete()
        except:
            pass
        self.__config.remove_section(server_uuid)
        
        store = _UscStore(self.__store_path, server_uuid)
        store.remove()

    def remove_server(self, server_path):
        """Removes a media server from the controller's list.

        Also removes the server side synchronized file and folders.
        server_path: d-bus path for the media server
        """

        server = Device(server_path)
        server_uuid = server.get_prop('UDN')

        if self.__config.has_section(server_uuid):
            self.__remove_server(server_uuid)

        self.__write_config()

    def reset(self):
        """Removes all media servers from the controller's list
        
        Also removes the server side synchronized file and folders.
        """

        for name in self.__config.sections():
            if name != UscController.DATA_PATH_SECTION:
                self.__remove_server(name)

        self.__write_config()

if __name__ == '__main__':
    print u'An Upload Sync Controller sample app.'
    print
    controller = UscController()
    controller.servers()
    print
    print u'"controller" instance is ready for use.'
    print u'Type "help(UscController)" for more details and usage samples.'



# mediaconsole
#
# Copyright (C) 2012-2017 Intel Corporation. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms and conditions of the GNU Lesser General Public License,
# version 2.1, as published by the Free Software Foundation.
#
# This program is distributed in the hope it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
#
# Mark Ryan <mark.d.ryan@intel.com>
#

import dbus
import sys
import json

from xml.dom import minidom

def print_properties(props):
    print json.dumps(props, indent=4, sort_keys=True)

class MediaObject(object):

    def __init__(self, path):
        bus = dbus.SessionBus()
        self._propsIF = dbus.Interface(bus.get_object(
                'com.intel.dleyna-server', path),
                                        'org.freedesktop.DBus.Properties')
        self.__objIF = dbus.Interface(bus.get_object(
                'com.intel.dleyna-server', path),
                                      'org.gnome.UPnP.MediaObject2')

    def get_props(self, iface = ""):
        return self._propsIF.GetAll(iface)

    def get_prop(self, prop_name, iface = ""):
        return self._propsIF.Get(iface, prop_name)

    def print_prop(self, prop_name, iface = ""):
        print_properties(self._propsIF.Get(iface, prop_name))

    def print_props(self, iface = ""):
        print_properties(self._propsIF.GetAll(iface))

    def print_dms_id(self):
        path = self._propsIF.Get("", "Path")
        dms_id = path[path.rfind("/") + 1:]
        i = 0
        while i+1 < len(dms_id):
            num = dms_id[i] + dms_id[i+1]
            sys.stdout.write(unichr(int(num, 16)))
            i = i + 2
        print

    def print_metadata(self):
        metadata = self.__objIF.GetMetaData()
        print minidom.parseString(metadata).toprettyxml(indent=' '*4)

    def delete(self):
        return self.__objIF.Delete()

    def update(self, to_add_update, to_delete):
        return self.__objIF.Update(to_add_update, to_delete)

    def get_metadata(self):
        return self.__objIF.GetMetaData()

class Item(MediaObject):
    def __init__(self, path):
        MediaObject.__init__(self, path)
        bus = dbus.SessionBus()
        self._itemIF = dbus.Interface(bus.get_object(
                'com.intel.dleyna-server', path),
                                       'org.gnome.UPnP.MediaItem2')

    def print_compatible_resource(self, protocol_info, fltr):
        print_properties(self._itemIF.GetCompatibleResource(protocol_info,
                                                             fltr))

class Container(MediaObject):

    def __init__(self, path):
        MediaObject.__init__(self, path)
        bus = dbus.SessionBus()
        self._containerIF = dbus.Interface(bus.get_object(
                'com.intel.dleyna-server', path),
                                            'org.gnome.UPnP.MediaContainer2')

    def list_children(self, offset, count, fltr, sort=""):
        objects = self._containerIF.ListChildrenEx(offset, count, fltr, sort)
        for item in objects:
            print_properties(item)
            print ""

    def list_containers(self, offset, count, fltr, sort=""):
        objects = self._containerIF.ListContainersEx(offset, count, fltr, sort)
        for item in objects:
            print_properties(item)
            print ""

    def list_items(self, offset, count, fltr, sort=""):
        objects = self._containerIF.ListItemsEx(offset, count, fltr, sort)
        for item in objects:
            print_properties(item)
            print ""

    def search(self, query, offset, count, fltr, sort=""):
        objects, total = self._containerIF.SearchObjectsEx(query, offset,
                                                            count, fltr, sort)
        print "Total Items: " + str(total)
        print
        for item in objects:
            print_properties(item)
            print ""

    def tree(self, level=0):
        objects = self._containerIF.ListChildren(
            0, 0, ["DisplayName", "Path", "Type"])
        for props in objects:
            print (" " * (level * 4) + props["DisplayName"] +
                   " : (" + props["Path"]+ ")")
            if props["Type"] == "container":
                Container(props["Path"]).tree(level + 1)

    def upload(self, name, file_path):
        (tid, path) = self._containerIF.Upload(name, file_path)
        print "Transfer ID: " + str(tid)
        print u"Path: " + path

    def create_container(self, name, type, child_types):
        path = self._containerIF.CreateContainer(name, type, child_types)
        print u"New container path: " + path

    def print_compatible_resource(self, protocol_info, fltr):
        print_properties(self._containerIF.GetCompatibleResource(protocol_info,
								 fltr))
    def create_reference(self, file_path):
        path = self._containerIF.CreateReference(file_path)
        print u"Reference Path: " + path

class Device(Container):

    def __init__(self, path):
        Container.__init__(self, path)
        bus = dbus.SessionBus()
        self._deviceIF = dbus.Interface(bus.get_object(
                'com.intel.dleyna-server', path),
                                         'com.intel.dLeynaServer.MediaDevice')

    def browse_objects(self, paths, fltr=['*']):
        objects = self._deviceIF.BrowseObjects(paths, fltr)
        for item in objects:
            print_properties(item)
            print ""

    def upload_to_any(self, name, file_path):
        (tid, path) = self._deviceIF.UploadToAnyContainer(name, file_path)
        print "Transfer ID: " + str(tid)
        print u"Path: " + path

    def create_container_in_any(self, name, type, child_types):
        path = self._deviceIF.CreateContainerInAnyContainer(name, type,
                                                             child_types)
        print u"New container path: " + path

    def get_upload_status(self, id):
        (status, length, total) = self._deviceIF.GetUploadStatus(id)
        print "Status: " + status
        print "Length: " + str(length)
        print "Total: " + str(total)

    def get_upload_ids(self):
        upload_ids  = self._deviceIF.GetUploadIDs()
        print_properties(upload_ids)

    def cancel_upload(self, id):
        self._deviceIF.CancelUpload(id)

    def cancel(self):
        return self._deviceIF.Cancel()

    def print_icon(self, mime_type, resolution):
        bytes, mime = self._deviceIF.GetIcon(mime_type, resolution)
        print "Icon mime type: " + mime

    def wake(self):
        return self._deviceIF.Wake()

class UPNP(object):

    def __init__(self):
        bus = dbus.SessionBus()
        self._manager = dbus.Interface(bus.get_object(
						'com.intel.dleyna-server',
						'/com/intel/dLeynaServer'),
				       'com.intel.dLeynaServer.Manager')
        self._propsIF = dbus.Interface(bus.get_object(
						'com.intel.dleyna-server',
						'/com/intel/dLeynaServer'),
				       'org.freedesktop.DBus.Properties')

    def get_props(self, iface = ""):
        return self._propsIF.GetAll(iface)

    def get_prop(self, prop_name, iface = ""):
        return self._propsIF.Get(iface, prop_name)

    def set_prop(self, prop_name, val, iface = ""):
        return self._propsIF.Set(iface, prop_name, val)

    def print_prop(self, prop_name, iface = ""):
        print_properties(self._propsIF.Get(iface, prop_name))

    def print_props(self, iface = ""):
        print_properties(self._propsIF.GetAll(iface))

    def server_from_name(self, friendly_name):
        retval = None
        for i in self._manager.GetServers():
            server = Device(i)
            server_name = server.get_prop("FriendlyName").lower()
            if server_name.find(friendly_name.lower()) != -1:
                retval = server
                break
        return retval

    def server_from_udn(self, udn):
        retval = None
        for i in self._manager.GetServers():
            server = Device(i)
            if server.get_prop("UDN") == udn:
                retval = server
                break
        return retval

    def servers(self):
        for i in self._manager.GetServers():
            try:
                server = Container(i)
                try:
                    folderName = server.get_prop("FriendlyName");
                except Exception:
                    folderName = server.get_prop("DisplayName");
                print u'{0:<30}{1:<30}'.format(folderName , i)
            except dbus.exceptions.DBusException, err:
                print u"Cannot retrieve properties for " + i
                print str(err).strip()[:-1]

    def version(self):
        print self._manager.GetVersion()

    def set_protocol_info(self, protocol_info):
        self._manager.SetProtocolInfo(protocol_info)

    def prefer_local_addresses(self, prefer):
        self._manager.PreferLocalAddresses(prefer)

    def rescan(self):
        self._manager.Rescan()

    def white_list_enable(self, enable):
        self.set_prop("WhiteListEnabled", enable)

    def white_list_add(self, entries):
	white_list = set(self.get_prop('WhiteListEntries'))
	white_list = white_list | set(entries)
        self.set_prop("WhiteListEntries", list(white_list))

    def white_list_remove(self, entries):
	white_list = set(self.get_prop('WhiteListEntries'))
	white_list = white_list - set(entries)
        self.set_prop("WhiteListEntries", list(white_list))

    def white_list_clear(self):
        self.set_prop("WhiteListEntries", [''])



# monitor_contents_changes
#
# Copyright (C) 2012-2017 Intel Corporation. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms and conditions of the GNU Lesser General Public License,
# version 2.1, as published by the Free Software Foundation.
#
# This program is distributed in the hope it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
#
# Regis Merlino <regis.merlino@intel.com>
#

import gobject
import dbus
import dbus.mainloop.glib
import json

def print_properties(props):
    print json.dumps(props, indent=4, sort_keys=True)

def properties_changed(iface, changed, invalidated, path):
    print "PropertiesChanged signal from {%s} [%s]" % (iface, path)
    print "Changed:"
    print_properties(changed)
    print "Invalidated:"
    print_properties(invalidated)

def container_update(id_list, path, interface):
    iface = interface[interface.rfind(".") + 1:]
    print "ContainerUpdateIDs signal from {%s} [%s]" % (iface, path)
    for (id_path, sys_id) in id_list:
        print "-->\t %s : %u" % (id_path, sys_id)

if __name__ == '__main__':
    dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)

    bus = dbus.SessionBus()

    bus.add_signal_receiver(properties_changed,
                            bus_name="com.intel.dleyna-server",
                            signal_name = "PropertiesChanged",
                            path_keyword="path")
    bus.add_signal_receiver(container_update,
                            bus_name="com.intel.dleyna-server",
                            signal_name = "ContainerUpdateIDs",
                            path_keyword="path",
                            interface_keyword="interface")

    mainloop = gobject.MainLoop()
    mainloop.run()



#!/usr/bin/env python

# monitor_last_change
#
# Copyright (C) 2012-2017 Intel Corporation. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms and conditions of the GNU Lesser General Public License,
# version 2.1, as published by the Free Software Foundation.
#
# This program is distributed in the hope it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
#
# Regis Merlino <regis.merlino@intel.com>
# Mark Ryan <mark.d.ryan@intel.com>
#

import gobject
import dbus
import dbus.mainloop.glib
import json

def print_properties(props):
    print json.dumps(props, indent=4, sort_keys=True)

def changed(objects, path):
    print "Changed signal from [%s]" % path
    print "Objects:"
    print_properties(objects)

if __name__ == '__main__':
    dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)

    bus = dbus.SessionBus()

    bus.add_signal_receiver(changed,
                            bus_name="com.intel.dleyna-server",
                            signal_name = "Changed",
                            path_keyword="path")

    mainloop = gobject.MainLoop()
    mainloop.run()



# stress-test
#
# Copyright (C) 2012-2017 Intel Corporation. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms and conditions of the GNU Lesser General Public License,
# version 2.1, as published by the Free Software Foundation.
#
# This program is distributed in the hope it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
#
# Regis Merlino <regis.merlino@intel.com>
#

import gobject
import dbus
import dbus.mainloop.glib

def handle_browse_reply(objects):
    print "Total Items: " + str(len(objects))

def handle_error(err):
    if err.get_dbus_name() == 'com.intel.dleyna.Cancelled':
        print "Cancelled..."
    else:
        print "An error occured"
        print err
        loop.quit()

def make_async_calls():
    i = 0
    while i < 5:
        root.ListChildrenEx(0,
                            5,
                            ["DisplayName", "Type"],
                            "-DisplayName",
                            reply_handler=handle_browse_reply,
                            error_handler=handle_error)
        i += 1
    device.Cancel()
    gobject.timeout_add(1000, make_async_calls)

dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)

bus = dbus.SessionBus()
root = dbus.Interface(bus.get_object(
        'com.intel.dleyna-server',
        '/com/intel/dLeynaServer/server/0'),
                      'org.gnome.UPnP.MediaContainer2')
device = dbus.Interface(bus.get_object(
        'com.intel.dleyna-server',
        '/com/intel/dLeynaServer/server/0'),
                        'com.intel.dLeynaServer.MediaDevice')

gobject.timeout_add(1000, make_async_calls)

loop = gobject.MainLoop()
loop.run()



# download_sync_controller
#
# Copyright (C) 2012-2017 Intel Corporation. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms and conditions of the GNU Lesser General Public License,
# version 2.1, as published by the Free Software Foundation.
#
# This program is distributed in the hope it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
#
# Regis Merlino <regis.merlino@intel.com>
#

import ConfigParser
import os
import shutil
import urllib
import dbus

from mediaconsole import UPNP, Container, Device

class _DscUpnp(UPNP):
    def __init__(self):
        UPNP.__init__(self)

    def get_servers(self):
        return self._manager.GetServers()

class _DscContainer(Container):
    def __init__(self, path):
        Container.__init__(self, path)
        self.__path = path

    def find_containers(self):
        return self._containerIF.SearchObjectsEx(
                                        'Type derivedfrom "container"',
                                        0, 0,
                                        ['DisplayName', 'Path', 'Type'], '')[0]
    def find_updates(self, upd_id):
        return self._containerIF.SearchObjectsEx(
                            'ObjectUpdateID > "{0}"'.format(upd_id),
                            0, 0,
                            ['DisplayName', 'Path', 'RefPath', 'URLs',
                             'Type', 'Parent'], '')[0]

    def find_children(self):
        return self._containerIF.ListChildrenEx(0, 0,
                                             ['DisplayName', 'Path', 'RefPath',
                                              'URLs', 'Parent', 'Type'], '')

class DscError(Exception):
    """A Download Sync Controller error."""
    def __init__(self, message):
        """
        message: description of the error

        """
        Exception.__init__(self, message)
        self.message = message

    def __str__(self):
        return 'DscError: ' + self.message

class _DscDownloader(object):
    def __init__(self, url, path):
        self.url = url
        self.path = path

    def download(self):
        urllib.urlretrieve(self.url, self.path)

class _DscStore(object):
    SYNC_SECTION = 'sync_info'
    SYNC_OPTION = 'sync_contents'
    CUR_ID_OPTION = 'current_id'
    MEDIA_SECTION = 'media_info'
    
    NAME_SUFFIX = '-name'

    ITEM_NEW = 1
    ITEM_UPDATE = 2
    CONTAINER_NEW = 3

    def __init__(self, root_path, server_id):
        self.__root_path = root_path + '/' + server_id
        self.__config_path = self.__root_path + '/' + 'tracking.conf'

        self.__config = ConfigParser.ConfigParser()
        self.__cur_id = 0
        self.__sync = False

    def initialize(self, sync):
        if not os.path.exists(self.__root_path):
            os.makedirs(self.__root_path)

        self.__config.read(self.__config_path)

        if not self.__config.has_section(_DscStore.SYNC_SECTION):
            self.__config.add_section(_DscStore.SYNC_SECTION)
            self.__config.set(_DscStore.SYNC_SECTION,
                              _DscStore.CUR_ID_OPTION, '0')
            if sync:
                self.__config.set(_DscStore.SYNC_SECTION,
                                  _DscStore.SYNC_OPTION, 'yes')
            else:
                self.__config.set(_DscStore.SYNC_SECTION,
                                  _DscStore.SYNC_OPTION, 'no')


        if not self.__config.has_section(_DscStore.MEDIA_SECTION):
            self.__config.add_section(_DscStore.MEDIA_SECTION)

        self.__cur_id = self.__config.getint(_DscStore.SYNC_SECTION,
                                             _DscStore.CUR_ID_OPTION)

        self.__sync = self.__config.getboolean(_DscStore.SYNC_SECTION,
                                               _DscStore.SYNC_OPTION)

    def __write_config(self):
        with open(self.__config_path, 'wb') as configfile:
            self.__config.write(configfile)

    def __id_from_path(self, path):
        return os.path.basename(path)

    def __orig_id(self, media_object):
        try:
            return self.__id_from_path(media_object['RefPath'])
        except KeyError:
            return self.__id_from_path(media_object['Path'])

    def __removed_items(self, local_ids, remote_items):
        for local_id in local_ids:
            found = False

            for remote in remote_items:
                remote_id = self.__id_from_path(remote['Path'])
                if local_id.endswith(_DscStore.NAME_SUFFIX) or \
                                                        local_id == remote_id:
                    found = True

            if not found:
                yield local_id

    def __sync_item(self, obj, obj_id, parent_id, status, write_conf):
        orig = self.__orig_id(obj)

        if status == _DscStore.ITEM_UPDATE:
            old_path = self.__config.get(_DscStore.MEDIA_SECTION, orig)
            new_path = self.__create_path_for_name(obj['DisplayName'])
            print u'\tMedia "{0}" updated'.format(obj['DisplayName'])
            print u'\t\tto "{0}"'.format(new_path)
            self.__config.set(_DscStore.MEDIA_SECTION, orig, new_path)
            os.rename(old_path, new_path)
        elif status == _DscStore.ITEM_NEW:
            print u'\tNew media "{0}" tracked'.format(obj['DisplayName'])
            self.__config.set(parent_id, obj_id, orig)
            self.__config.set(parent_id, obj_id + _DscStore.NAME_SUFFIX,
                              obj['DisplayName'])
            if not self.__config.has_option(_DscStore.MEDIA_SECTION, orig) and \
                                                                    self.__sync:
                local_path = self.__create_path_for_name(obj['DisplayName'])
                self.__config.set(_DscStore.MEDIA_SECTION, orig, local_path)
                print u'\tDownloading contents from "{0}"'.format(obj['URLs'][0])
                print u'\t\tinto "{0}"...'.format(local_path)
                downloader = _DscDownloader(obj['URLs'][0], local_path)
                downloader.download()
        else:
            pass

        if write_conf:
            self.__write_config()

    def __create_path_for_name(self, file_name):
        new_path = self.__root_path + '/' + str(self.__cur_id) + '-' + file_name

        self.__cur_id += 1
        self.__config.set(_DscStore.SYNC_SECTION, _DscStore.CUR_ID_OPTION,
                          str(self.__cur_id))

        return new_path

    def remove(self):
        if os.path.exists(self.__root_path):
            shutil.rmtree(self.__root_path)

    def sync_container(self, container, items):
        print u'Syncing container "{0}"...'.format(container['DisplayName'])

        container_id = self.__id_from_path(container['Path'])
        if not self.__config.has_section(container_id):
            self.__config.add_section(container_id)

        for remote in items:
            remote_id = self.__id_from_path(remote['Path'])
            if not self.__config.has_option(container_id, remote_id):
                if remote['Type'] == 'container':
                    status = _DscStore.CONTAINER_NEW
                else:
                    status = _DscStore.ITEM_NEW
                self.__sync_item(remote, remote_id, container_id, status, False)

        for local in self.__removed_items(
                                    self.__config.options(container_id), items):
            if self.__config.has_section(local):
                print u'\tRemoved a container'
                self.__config.remove_option(container_id, local)
                self.__config.remove_section(local)
            else:
                orig = self.__config.get(container_id, local)
                name = self.__config.get(container_id,
                                              local + _DscStore.NAME_SUFFIX)
                print u'\tRemoved media "{0}"'.format(name)
                self.__config.remove_option(container_id, local)
                self.__config.remove_option(container_id,
                                            local + _DscStore.NAME_SUFFIX)
                if local == orig:
                    orig_name = self.__config.get(_DscStore.MEDIA_SECTION, orig)
                    self.__config.remove_option(_DscStore.MEDIA_SECTION, orig)
                    if self.__sync:
                        print u'\tRemoved local downloaded contents "{0}"' \
                                                            .format(orig_name)
                        if os.path.exists(orig_name):
                            os.remove(orig_name)

        self.__write_config()

    def sync_item(self, obj):
        print u'Syncing item "{0}"...'.format(obj['DisplayName'])
        obj_id = self.__id_from_path(obj['Path'])
        parent_id = self.__id_from_path(obj['Parent'])
        if self.__config.has_option(parent_id, obj_id):
            status = _DscStore.ITEM_UPDATE
        else:
            status = _DscStore.ITEM_NEW

        self.__sync_item(obj, obj_id, parent_id, status, True)

class DscController(object):
    """A Download Sync Controller.

    The Download Sync Controller receive changes in the content or metadata
    stored on media servers (DMS/M-DMS) and apply those changes to
    the local storage.
    Media servers must expose the 'content-synchronization' capability to
    be tracked by this controller.

    The three main methods are servers(), track() and sync().
    * servers() lists the media servers available on the network
    * track() is used to add a media server to the list of servers that are
      to be synchronized.
    * sync() launches the servers synchronisation to a local storage

    Sample usage:
    >>> controller.servers()
    >>> controller.track('/com/intel/dLeynaServer/server/0')
    >>> controller.sync()

    """
    CONFIG_PATH = os.environ['HOME'] + '/.config/download-sync-controller.conf'
    SUID_OPTION = 'system_update_id'
    SRT_OPTION = 'service_reset_token'
    SYNC_OPTION = 'sync_contents'
    DATA_PATH_SECTION = '__data_path__'
    DATA_PATH_OPTION = 'path'

    def __init__(self, rel_path = None):
        """
        rel_path: if provided, contains the relative local storage path,
                  from the user's HOME directory.
                  If not provided, the local storage path will be
                  '$HOME/download-sync-controller'

        """
        self.__upnp = _DscUpnp()

        self.__config = ConfigParser.ConfigParser()
        self.__config.read(DscController.CONFIG_PATH)
        
        if rel_path:
            self.__set_data_path(rel_path)
        elif not self.__config.has_section(DscController.DATA_PATH_SECTION):
            self.__set_data_path('download-sync-controller')
            
        self.__store_path = self.__config.get(DscController.DATA_PATH_SECTION,
                                              DscController.DATA_PATH_OPTION)

    def __write_config(self):
        with open(DscController.CONFIG_PATH, 'wb') as configfile:
            self.__config.write(configfile)

    def __set_data_path(self, rel_path):
        data_path = os.environ['HOME'] + '/' + rel_path

        if not self.__config.has_section(DscController.DATA_PATH_SECTION):
            self.__config.add_section(DscController.DATA_PATH_SECTION)

        self.__config.set(DscController.DATA_PATH_SECTION,
                          DscController.DATA_PATH_OPTION, data_path)

        self.__write_config()

    def __need_sync(self, servers):
        for item in servers:
            device = Device(item)
            uuid = device.get_prop('UDN')

            if self.__config.has_section(uuid):
                new_id = device.get_prop('SystemUpdateID')
                new_srt = device.get_prop('ServiceResetToken')
                cur_id = self.__config.getint(uuid, DscController.SUID_OPTION)
                cur_srt = self.__config.get(uuid, DscController.SRT_OPTION)
                if cur_id == -1 or cur_srt != new_srt:
                    print
                    print u'Server {0} needs *full* sync:'.format(uuid)
                    yield item, uuid, 0, new_id, new_srt, True
                elif cur_id < new_id:
                    print
                    print u'Server {0} needs sync:'.format(uuid)
                    yield item, uuid, cur_id, new_id, new_srt, False

    def __check_trackable(self, server):
        try:
            try:
                srt = server.get_prop('ServiceResetToken')
            except:
                raise DscError("'ServiceResetToken' variable not supported")

            try:
                dlna_caps = server.get_prop('DLNACaps')
                if not 'content-synchronization' in dlna_caps:
                    raise
            except:
                raise DscError("'content-synchronization' cap not supported")

            try:
                search_caps = server.get_prop('SearchCaps')
                if not [x for x in search_caps if 'ObjectUpdateID' in x]:
                    raise
                if not [x for x in search_caps if 'ContainerUpdateID' in x]:
                    raise
            except:
                raise DscError("'objectUpdateID' search cap not supported")

            return srt
        except DscError as err:
            print err
            return None
    
    def track(self, server_path, track = True, sync_contents = True):
        """Adds or removes a media server to/from the controller's list.

        server_path: d-bus path for the media server
        track: when 'True', adds a server to the list
               when 'False' removes a server from the list
        sync_contents: when 'True', downloads media contents to the local
                       storage upon synchronization.

        """
        server = Device(server_path)
        server_uuid = server.get_prop('UDN')
        
        if track and not self.__config.has_section(server_uuid):
            srt = self.__check_trackable(server)
            if srt != None:
                self.__config.add_section(server_uuid)

                self.__config.set(server_uuid, DscController.SUID_OPTION, '-1')
                self.__config.set(server_uuid, DscController.SRT_OPTION, srt)
                if sync_contents:
                    self.__config.set(server_uuid, DscController.SYNC_OPTION,
                                      'yes')
                else:
                    self.__config.set(server_uuid, DscController.SYNC_OPTION,
                                      'no')

                self.__write_config()
            else:
                print u"Sorry, the server {0} has no such capability and " \
                        "will not be tracked.".format(server_path)

        elif not track and self.__config.has_section(server_uuid):
            self.__config.remove_section(server_uuid)
            self.__write_config()
            
            store = _DscStore(self.__store_path, server_uuid)
            store.remove()

    def track_reset(self, server_path, sync_contents = True):
        """Removes local contents and meta data for a media server.

        The next synchronization will be a *full* synchronization.

        server_path: d-bus path for the media server
        sync_contents: when 'True', downloads media contents to the local
                       storage upon synchronization.

        """
        self.track(server_path, False, sync_contents)
        self.track(server_path, True, sync_contents)

    def servers(self):
        """Displays media servers available on the network.

        Displays media servers information as well as the tracked status.

        """
        print u'Running servers:'

        for item in self.__upnp.get_servers():
            try:
                server = Container(item)
                try:
                    folder_name = server.get_prop('FriendlyName')
                except Exception:
                    folder_name = server.get_prop('DisplayName')
                device = Device(item)
                dev_uuid = device.get_prop('UDN')
                dev_path = device.get_prop('Path')

                print u'{0:<25}  Tracked({2})  {3}  {1}'.format(folder_name,
                                            dev_path,
                                            self.__config.has_option(dev_uuid,
                                                    DscController.SUID_OPTION),
                                            dev_uuid)

            except dbus.exceptions.DBusException as err:
                print u'Cannot retrieve properties for ' + item
                print str(err).strip()[:-1]

    def tracked_servers(self):
        """Displays the list of servers currently tracked by the controller."""
        print u'Tracked servers:'

        for name in self.__config.sections():
            if name != DscController.DATA_PATH_SECTION:
                print u'{0:<30}'.format(name)

    def sync(self):
        """Performs a synchronization for all the tracked media servers.

        Displays some progress information during the process.

        """
        print u'Syncing...'

        for item, uuid, cur, new_id, new_srt, full_sync in \
                                    self.__need_sync(self.__upnp.get_servers()):
            sync = self.__config.getboolean(uuid, DscController.SYNC_OPTION)

            if full_sync:
                print u'Resetting local contents for server {0}'.format(uuid)

                self.track_reset(item)

                objects = _DscContainer(item).find_containers()
            else:
                objects = _DscContainer(item).find_updates(cur)

            store = _DscStore(self.__store_path, uuid)
            store.initialize(sync)

            for obj in objects:
                if obj['Type'] == 'container':
                    children = _DscContainer(obj['Path']).find_children()
                    store.sync_container(obj, children)
                else:
                    store.sync_item(obj)

            self.__config.set(uuid, DscController.SUID_OPTION, str(new_id))
            if full_sync:
                self.__config.set(uuid, DscController.SRT_OPTION, str(new_srt))
            self.__write_config()

        print
        print u'Done.'

    def reset(self):
        """Removes local contents and meta data for all the tracked servers.

        After the call, the list of tracked servers will be empty. 

        """
        for name in self.__config.sections():
            if name != DscController.DATA_PATH_SECTION:
                self.__config.remove_section(name)

                store = _DscStore(self.__store_path, name)
                store.remove()

        self.__write_config()


if __name__ == '__main__':
    controller = DscController()
    controller.servers()
    print
    print u'"controller" instance is ready for use.'
    print u'Type "help(DscController)" for more details and usage samples.'
